# Copyright 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli import CDP_ACCESS_KEY_ID_KEY_NAME, CDP_PRIVATE_KEY_KEY_NAME
from cdpcli.extensions.commands import BasicCommand
from cdpcli.extensions.configure import CREDENTIAL_FILE_COMMENT
from cdpcli.extensions.writer import ConfigFileWriter


class LogoutCommand(BasicCommand):
    """
    The 'cdp logout' command handler.
    """

    NAME = 'logout'
    DESCRIPTION = 'Logout CDP.'
    SYNOPSIS = 'cdp logout'
    EXAMPLES = (
        'To logout::\n'
        '\n'
        '    $ cdp logout\n'
    )
    SUBCOMMANDS = []
    ARG_TABLE = []

    def __init__(self, config_writer=None):
        super(LogoutCommand, self).__init__()
        if config_writer is None:
            config_writer = ConfigFileWriter()
        self._config_writer = config_writer

    def _run_main(self, client_creator, parsed_args, parsed_globals):
        # Called when invoked with no sub-command "cdp logout"
        context = client_creator.context
        self._clear_access_token(context)

    def _clear_access_token(self, context):
        credential_file_values = {CDP_ACCESS_KEY_ID_KEY_NAME: '',
                                  CDP_PRIVATE_KEY_KEY_NAME: ''}
        profile_name = context.effective_profile
        if profile_name is not None:
            credential_file_values['__section__'] = profile_name
        shared_credentials_filename = os.path.expanduser(
            context.get_config_variable('credentials_file'))
        self._config_writer.update_config(
            credential_file_values,
            shared_credentials_filename,
            config_file_comment=CREDENTIAL_FILE_COMMENT)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import setup_common
from setuptools import find_packages
from setuptools import setup
import versioneer


release = 'public'

setup(
    name='cdpcli',
    version=versioneer.get_version(),
    description='Cloudera CDP Command Line Interface',
    long_description=setup_common.read_long_description(),
    long_description_content_type='text/markdown',
    url='https://console.cdp.cloudera.com/',
    license='Apache License 2.0',
    author='Cloudera, Inc.',
    classifiers=setup_common.get_classifiers(release),
    packages=find_packages(exclude=['tests']),
    include_package_data=True,
    install_requires=setup_common.get_requirements(),
    entry_points=setup_common.get_entry_points(),
    cmdclass=versioneer.get_cmdclass(),
)



# Licensed to Cloudera, Inc. under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  Cloudera, Inc. licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from cdpcli.extensions.df import DfExtension
from cdpcli.extensions.df.createdeployment import CreateDeployment
from cdpcli.extensions.df.createdeployment import OPERATION_CLI_NAME \
    as CREATE_DEPLOYMENT_OPERATION_CLI_NAME


def register_extension(operation_callers, operation_model):
    """
    Register an extension to run before or after the CLI command.
    To replace the original CLI caller:
    * operation_callers.insert(0, ReplacementCaller())
    * return False by the ReplacementCaller.invoke(...)
    """
    operation_callers.insert(0, DfExtension())


def register_command(clidriver, service_model, command_table):
    """
    Register an additional command to run.
    """
    command_table[CREATE_DEPLOYMENT_OPERATION_CLI_NAME] = \
        CreateDeployment(clidriver, service_model)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli import model
from cdpcli import shorthand
from nose.tools import assert_equal
from tests import unittest


def test_parse():
    # Key val pairs with scalar value.
    yield (_can_parse, 'foo=bar', {'foo': 'bar'})
    yield (_can_parse, 'foo=bar', {'foo': 'bar'})
    yield (_can_parse, 'foo=bar,baz=qux', {'foo': 'bar', 'baz': 'qux'})
    yield (_can_parse, 'a=b,c=d,e=f', {'a': 'b', 'c': 'd', 'e': 'f'})
    # Empty values are allowed.
    yield (_can_parse, 'foo=', {'foo': ''})
    yield (_can_parse, 'foo=,bar=', {'foo': '', 'bar': ''})
    # Unicode is allowed.
    yield (_can_parse, u'foo=\u2713', {'foo': u'\u2713'})
    yield (_can_parse, u'foo=\u2713,\u2713', {'foo': [u'\u2713', u'\u2713']})
    # Key val pairs with csv values.
    yield (_can_parse, 'foo=a,b', {'foo': ['a', 'b']})
    yield (_can_parse, 'foo=a,b,c', {'foo': ['a', 'b', 'c']})
    yield (_can_parse, 'foo=a,b,bar=c,d', {'foo': ['a', 'b'],
                                           'bar': ['c', 'd']})
    yield (_can_parse, 'foo=a,b,c,bar=d,e,f',
           {'foo': ['a', 'b', 'c'], 'bar': ['d', 'e', 'f']})
    # Spaces in values are allowed.
    yield (_can_parse, 'foo=a,b=with space', {'foo': 'a', 'b': 'with space'})
    # Trailing spaces are still ignored.
    yield (_can_parse, 'foo=a,b=with trailing space  ',
           {'foo': 'a', 'b': 'with trailing space'})
    yield (_can_parse, 'foo=first space',
           {'foo': 'first space'})
    yield (_can_parse, 'foo=a space,bar=a space,baz=a space',
           {'foo': 'a space', 'bar': 'a space', 'baz': 'a space'})

    # Dashes are allowed in key names.
    yield (_can_parse, 'with-dash=bar', {'with-dash': 'bar'})

    # Underscore are also allowed.
    yield (_can_parse, 'with_underscore=bar', {'with_underscore': 'bar'})

    # Dots are allowed.
    yield (_can_parse, 'with.dot=bar', {'with.dot': 'bar'})

    # Explicit lists.
    yield (_can_parse, 'foo=[]', {'foo': []})
    yield (_can_parse, 'foo=[a]', {'foo': ['a']})
    yield (_can_parse, 'foo=[a,b]', {'foo': ['a', 'b']})
    yield (_can_parse, 'foo=[a,b,c]', {'foo': ['a', 'b', 'c']})
    yield (_can_parse, 'foo=[a,b],bar=c,d',
           {'foo': ['a', 'b'], 'bar': ['c', 'd']})
    yield (_can_parse, 'foo=[a,b],bar=[c,d]',
           {'foo': ['a', 'b'], 'bar': ['c', 'd']})
    yield (_can_parse, 'foo=a,b,bar=[c,d]',
           {'foo': ['a', 'b'], 'bar': ['c', 'd']})
    yield (_can_parse, 'foo=[a=b,c=d]', {'foo': ['a=b', 'c=d']})
    yield (_can_parse, 'foo=[a=b,c=d]', {'foo': ['a=b', 'c=d']})
    # Lists with whitespace.
    yield (_can_parse, 'foo=[ a , b  , c  ]',
           {'foo': ['a', 'b', 'c']})
    yield (_can_parse, 'foo  =  [ a , b  , c  ]',
           {'foo': ['a', 'b', 'c']})
    yield (_can_parse, 'foo=[,,]', {'foo': ['', '']})

    # Single quoted strings.
    yield (_can_parse, "foo='bar'", {"foo": "bar"})
    yield (_can_parse, "foo='bar,baz'", {"foo": "bar,baz"})
    # Single quoted strings for each value in a CSV list.
    yield (_can_parse, "foo='bar','baz'", {"foo": ['bar', 'baz']})
    # Can mix single quoted and non quoted values.
    yield (_can_parse, "foo=bar,'baz'", {"foo": ['bar', 'baz']})
    # Quoted strings can include chars not allowed in unquoted strings.
    yield (_can_parse, "foo=bar,'baz=qux'", {"foo": ['bar', 'baz=qux']})
    yield (_can_parse, "foo=bar,'--option=bar space'",
           {"foo": ['bar', '--option=bar space']})
    # Can escape the single quote.
    yield (_can_parse, "foo='bar\\'baz'", {"foo": "bar'baz"})
    yield (_can_parse, "foo='bar\\\\baz'", {"foo": "bar\\baz"})

    # Double quoted strings.
    yield (_can_parse, 'foo="bar"', {'foo': 'bar'})
    yield (_can_parse, 'foo="bar,baz"', {'foo': 'bar,baz'})
    yield (_can_parse, 'foo="bar","baz"', {'foo': ['bar', 'baz']})
    yield (_can_parse, 'foo=bar,"baz=qux"', {'foo': ['bar', 'baz=qux']})
    yield (_can_parse, 'foo=bar,"--option=bar space"',
           {'foo': ['bar', '--option=bar space']})
    yield (_can_parse, 'foo="bar\\"baz"', {'foo': 'bar"baz'})
    yield (_can_parse, 'foo="bar\\\\baz"', {'foo': 'bar\\baz'})

    # Can escape comma in CSV list.
    yield (_can_parse, 'foo=a\\,b', {"foo": "a,b"})
    yield (_can_parse, 'foo=a\\,b', {"foo": "a,b"})
    yield (_can_parse, 'foo=a\\,', {"foo": "a,"})
    yield (_can_parse, 'foo=\\,', {"foo": ","})
    yield (_can_parse, 'foo=a,b\\,c', {"foo": ['a', 'b,c']})
    yield (_can_parse, 'foo=a,b\\,', {"foo": ['a', 'b,']})
    yield (_can_parse, 'foo=a,\\,bc', {"foo": ['a', ',bc']})

    # Ignores whitespace around '=' and ','
    yield (_can_parse, 'foo= bar', {'foo': 'bar'})
    yield (_can_parse, 'foo =bar', {'foo': 'bar'})
    yield (_can_parse, 'foo = bar', {'foo': 'bar'})
    yield (_can_parse, 'foo  =   bar', {'foo': 'bar'})
    yield (_can_parse, 'foo = bar,baz = qux', {'foo': 'bar', 'baz': 'qux'})
    yield (_can_parse, 'a = b,  c = d , e = f', {'a': 'b', 'c': 'd', 'e': 'f'})
    yield (_can_parse, 'foo = ', {'foo': ''})
    yield (_can_parse, 'a=b,c=  d,  e,  f', {'a': 'b', 'c': ['d', 'e', 'f']})
    yield (_can_parse, 'Name=foo,Values=  a  ,  b  ,  c  ',
           {'Name': 'foo', 'Values': ['a', 'b', 'c']})
    yield (_can_parse, 'Name=foo,Values= a,  b  ,  c',
           {'Name': 'foo', 'Values': ['a', 'b', 'c']})

    # Can handle newlines between values.
    yield (_can_parse, 'Name=foo,\nValues=a,b,c',
           {'Name': 'foo', 'Values': ['a', 'b', 'c']})
    yield (_can_parse, 'A=b,\nC=d,\nE=f\n',
           {'A': 'b', 'C': 'd', 'E': 'f'})

    # Hashes
    yield (_can_parse, 'Name={foo=bar,baz=qux}',
           {'Name': {'foo': 'bar', 'baz': 'qux'}})
    yield (_can_parse, 'Name={foo=[a,b,c],bar=baz}',
           {'Name': {'foo': ['a', 'b', 'c'], 'bar': 'baz'}})
    yield (_can_parse, 'Name={foo=bar},Bar=baz',
           {'Name': {'foo': 'bar'}, 'Bar': 'baz'})
    yield (_can_parse, 'Bar=baz,Name={foo=bar}',
           {'Bar': 'baz', 'Name': {'foo': 'bar'}})
    yield (_can_parse, 'a={b={c=d}}',
           {'a': {'b': {'c': 'd'}}})
    yield (_can_parse, 'a={b={c=d,e=f},g=h}',
           {'a': {'b': {'c': 'd', 'e': 'f'}, 'g': 'h'}})

    # Combining lists and hashes.
    yield (_can_parse, 'Name=[{foo=bar}, {baz=qux}]',
           {'Name': [{'foo': 'bar'}, {'baz': 'qux'}]})

    # Combining hashes and lists.
    yield (_can_parse, 'Name=[{foo=[a,b]}, {bar=[c,d]}]',
           {'Name': [{'foo': ['a', 'b']}, {'bar': ['c', 'd']}]})


def test_error_parsing():
    yield (_is_error, 'foo')
    # Missing closing quotes
    yield (_is_error, 'foo="bar')
    yield (_is_error, "foo='bar")
    yield (_is_error, "foo=[bar")
    yield (_is_error, "foo={bar")
    yield (_is_error, "foo={bar}")
    yield (_is_error, "foo={bar=bar")
    yield (_is_error, "foo=bar,")
    yield (_is_error, "foo==bar,\nbar=baz")


def _is_error(expr):
    try:
        shorthand.ShorthandParser().parse(expr)
    except shorthand.ShorthandParseError:
        pass
    except Exception as e:
        raise AssertionError(
            "Expected ShorthandParseError, but received unexpected "
            "exception instead (%s): %s" % (e.__class__, e))
    else:
        raise AssertionError("Expected ShorthandParseError, but no "
                             "exception was raised for expression: %s" % expr)


def _can_parse(data, expected):
    actual = shorthand.ShorthandParser().parse(data)
    assert_equal(actual, expected)


class TestModelVisitor(unittest.TestCase):
    def test_promote_to_list_of_ints(self):
        m = model.DenormalizedStructureBuilder().with_members({
            'A': {
                'type': 'array',
                'items': {'type': 'string'}
            },
        }).build_model()
        b = shorthand.BackCompatVisitor()

        params = {'A': 'foo'}
        b.visit(params, m)
        self.assertEqual(params, {'A': ['foo']})

    def test_dont_promote_list_if_none_value(self):
        m = model.DenormalizedStructureBuilder().with_members({
            'A': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'Single': {'type': 'string'}
                    },
                },
            },
        }).build_model()
        b = shorthand.BackCompatVisitor()
        params = {}
        b.visit(params, m)
        self.assertEqual(params, {})

    def test_can_convert_scalar_types_from_string(self):
        m = model.DenormalizedStructureBuilder().with_members({
            'A': {'type': 'integer'},
            'B': {'type': 'string'},
            'C': {'type': 'float'},
            'D': {'type': 'boolean'},
            'E': {'type': 'boolean'},
            'F': {'type': 'blob'},
        }).build_model()
        b = shorthand.BackCompatVisitor()

        params = {'A': '24', 'B': '24', 'C': '24.12345',
                  'D': 'true', 'E': 'false',
                  'F': 'SGVsbG8gV29ybGQ='}
        b.visit(params, m)
        self.assertEqual(
            params,
            {'A': 24, 'B': '24', 'C': float('24.12345'),
             'D': True, 'E': False,
             'F': 'SGVsbG8gV29ybGQ='})

    def test_empty_values_not_added(self):
        m = model.DenormalizedStructureBuilder().with_members({
            'A': {'type': 'boolean'},
        }).build_model()
        b = shorthand.BackCompatVisitor()

        params = {}
        b.visit(params, m)
        self.assertEqual(params, {})

    def test_can_convert_list_of_integers(self):
        m = model.DenormalizedStructureBuilder().with_members({
            'A': {
                'type': 'array',
                'items': {
                    'type': 'integer',
                },
            },
        }).build_model()
        b = shorthand.BackCompatVisitor()
        params = {'A': ['1', '2']}
        b.visit(params, m)
        # We should have converted each list element to an integer
        # because the type of the list member is integer.
        self.assertEqual(params, {'A': [1, 2]})

    def test_can_convert_map_of_integers(self):
        m = model.DenormalizedStructureBuilder().with_members({
            'A': {
                'type': 'object',
                'additionalProperties': {
                    'type': 'integer'
                },
            },
        }).build_model()
        b = shorthand.BackCompatVisitor()
        params = {'A': {'key': '1'}}
        b.visit(params, m)
        # We should have converted each map value to an integer
        # because the type of the value member is integer.
        self.assertEqual(params, {'A': {'key': 1}})



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016-2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
import pprint
import threading
import time

from cdpcli.cdprequest import create_request_object
from cdpcli.compat import six
from cdpcli.exceptions import EndpointConnectionError
from cdpcli.parser import ResponseParserFactory
from requests.exceptions import ConnectionError
from requests.sessions import Session
from requests.utils import get_environ_proxies


DEFAULT_TIMEOUT = 60
LOG = logging.getLogger('cdpcli.endpoint')


def convert_to_response_dict(http_response, operation_model):
    response_dict = {
        'headers': http_response.headers,
        'status_code': http_response.status_code,
    }
    if response_dict['status_code'] >= 300:
        response_dict['body'] = http_response.content
    else:
        response_dict['body'] = http_response.content
    return response_dict


class PreserveAuthSession(Session):

    def rebuild_auth(self, prepared_request, response):
        pass


class EndpointResolver(object):
    """A builder of endpoint URLs for Altus and CDP services. Used by EndpointCreator."""
    ENDPOINT_URL_KEY_NAME = 'endpoint_url'
    CDP_ENDPOINT_URL_KEY_NAME = 'cdp_endpoint_url'
    LOGIN_URL_KEY_NAME = 'login_url'
    CDP_REGION_KEY_NAME = 'cdp_region'

    def _construct_altus_endpoint(self, service_name, scheme, region, port):
        """Construct a default base URL to an Altus service.

        :param service_name: service name, as referenced in its URLs
        :param scheme: URL scheme, e.g., 'https'
        :param port: service port
        :param region: service region
        :returns: Altus service base URL
        """
        if region in ['default', 'us-west-1']:
            return "%s://%sapi.us-west-1.altus.cloudera.com:%d" % \
                   (scheme, service_name, port)
        else:
            return "%s://api.%s.cdp.cloudera.com:%d" % (scheme, region, port)

    def _construct_cdp_endpoint(self, scheme, prefix, region, port):
        """Construct a default base URL to a CDP service.

        :param scheme: URL scheme, e.g., 'https'
        :param prefix: CDP API prefix used in URLs, e.g., 'api'
        :param port: service port
        :param region: service region
        :returns: CDP service base URL
        """
        if region == 'default':
            region = 'us-west-1'
        return "%s://%s.%s.cdp.cloudera.com:%d" % (scheme, prefix, region, port)

    def _construct_login_endpoint(self, scheme, region, port):
        """Construct a login URL to a CDP service.

        :param scheme: URL scheme, e.g., 'https'
        :param port: service port
        :param region: service region
        :returns: CDP login URL
        """
        if region == 'default':
            region = 'us-west-1'
        if region == 'us-west-1':
            return "%s://consoleauth.altus.cloudera.com:%d/login" % (scheme, port)
        return "%s://console.%s.cdp.cloudera.com:%d/consoleauth/login" % \
               (scheme, region, port)

    def _substitute_custom_endpoint(self, endpoint_url, service_name, prefix, products):
        """Applies string formatting for one %s value.

        :param endpoint_url: URL format string, either no %s or contains one %s
        :param service_name: Altus service name, as referenced in its URLs.
                             Or 'LOGIN' for login command
        :param prefix: CDP API prefix used in URLs, e.g., 'api'
        :param products: service products (ALTUS or CDP)
        :returns: formatted string
        """
        # For an explicit endpoint URL, swap in the service name (Altus) or prefix
        # (CDP) if there is a spot for it, and then return it.
        if endpoint_url.count('%s') == 1:
            if products == ['CDP']:
                return endpoint_url % prefix
            else:
                return endpoint_url % service_name
        else:
            return endpoint_url

    def _read_region_from_config(self, explicit_region, config):
        """Reads region from config if it is not explicitly specified.

        :param explicit_region: The explicitly specified region
        :param config: CLI configuration
        :return:
        """
        if explicit_region is None:
            region = 'default'
        else:
            region = explicit_region
        if region == 'default':
            region_from_config = config.get(EndpointResolver.CDP_REGION_KEY_NAME)
            if region_from_config is not None:
                region = region_from_config
        return region

    def resolve(self, service_name, prefix, products, explicit_endpoint_url, config,
                region, scheme, port):
        """Creates a fully-resolved URL for a service endpoint.

        :param service_name: Altus service name, as referenced in its URLs.
                             Or 'LOGIN' for login command
        :param prefix: CDP API prefix used in URLs, e.g., 'api'
        :param products: service products (ALTUS or CDP)
        :param explicit_endpoint_url: explicit endpoint URL which overrides any other
        :param config: CLI configuration
        :param region: service region
        :param scheme: URL scheme, e.g., 'https'
        :param port: service port
        :returns: resolved URL for service endpoint
        """
        if explicit_endpoint_url is not None:
            return self._substitute_custom_endpoint(explicit_endpoint_url,
                                                    service_name, prefix, products)

        if service_name == 'LOGIN':
            # Login endpoint
            endpoint_from_config = config.get(EndpointResolver.LOGIN_URL_KEY_NAME)
            if endpoint_from_config is not None:
                return endpoint_from_config
            region = self._read_region_from_config(region, config)
            return self._construct_login_endpoint(scheme, region, port)

        if products == ['CDP']:
            # If a CDP endpoint base URL has been configured, use it, swapping in the
            # prefix if there is a spot for it. Otherwise, construct a default URL for
            # the prefix.
            endpoint_from_config = config.get(EndpointResolver.CDP_ENDPOINT_URL_KEY_NAME)
            if endpoint_from_config is not None:
                return self._substitute_custom_endpoint(endpoint_from_config,
                                                        service_name, prefix, products)
            region = self._read_region_from_config(region, config)
            return self._construct_cdp_endpoint(scheme, prefix, region, port)
        else:
            # If an Altus endpoint base URL has been configured, use it, swapping in the
            # prefix if there is a spot for it. Otherwise, construct a default URL for
            # the prefix.
            endpoint_from_config = config.get(EndpointResolver.ENDPOINT_URL_KEY_NAME)
            if endpoint_from_config is not None:
                return self._substitute_custom_endpoint(endpoint_from_config,
                                                        service_name, prefix, products)
            region = self._read_region_from_config(region, config)
            return self._construct_altus_endpoint(service_name, scheme, region, port)


class EndpointCreator(object):
    """A creator of Endpoints for Altus and CDP services."""

    def __init__(self, endpoint_resolver):
        """Construct an endpoint creator.

        :param endpoint_resolver: Endpoint resolver to use
        """
        self._endpoint_resolver = endpoint_resolver

    def create_endpoint(self,
                        service_model,
                        explicit_endpoint_url,
                        scoped_config,
                        region,
                        response_parser_factory,
                        tls_verification,
                        timeout,
                        retry_handler):
        """Create an Endpoint for a service. Uses scheme 'https' and port 443 for any
           non-explicit endpoint URL. The returned endpoint uses proxies set in the
           environment, as understood by the requests library.

        :param service_model: ServiceModel holding service information
        :param explicit_endpoint_url: explicit endpoint URL which overrides any other
        :param scoped_config: CLI configuration
        :param region: service region
        :param response_parser_factory: set in endpoint
        :param tls_verification: set in endpoint
        :param timeout: set in endpoint
        :param retry_handler: set in endpoint
        :returns: service endpoint
        """
        endpoint_url = \
            self._endpoint_resolver.resolve(explicit_endpoint_url=explicit_endpoint_url,
                                            config=scoped_config,
                                            region=region,
                                            service_name=service_model.endpoint_name,
                                            prefix=service_model.endpoint_prefix,
                                            products=service_model.products,
                                            scheme='https',
                                            port=443)
        proxies = self._get_proxies(endpoint_url)
        return Endpoint(
            endpoint_url,
            proxies=proxies,
            tls_verification=tls_verification,
            timeout=timeout,
            response_parser_factory=response_parser_factory,
            retry_handler=retry_handler)

    def _get_proxies(self, url):
        return get_environ_proxies(url)


class Endpoint(object):

    def __init__(self,
                 host,
                 proxies=None,
                 tls_verification=True,
                 timeout=DEFAULT_TIMEOUT,
                 response_parser_factory=None,
                 retry_handler=None):
        self.host = host
        self.tls_verification = tls_verification
        if proxies is None:
            proxies = {}
        self.proxies = proxies
        self.http_session = PreserveAuthSession()
        self.timeout = timeout
        self._lock = threading.Lock()
        if response_parser_factory is None:
            response_parser_factory = ResponseParserFactory()
        self._response_parser_factory = response_parser_factory
        self._retry_handler = retry_handler

    def __repr__(self):
        return '%s' % (self.host)

    def make_request(self, operation_model, request_dict, request_signer,
                     allow_redirects=True):
        return self._send_request(request_dict, operation_model, request_signer,
                                  allow_redirects)

    def create_request(self, params, operation_model, request_signer):
        request = create_request_object(params)
        request_signer.sign(request)
        return self.prepare_request(request)

    def _encode_headers(self, headers):
        # In place encoding of headers to utf-8 if they are unicode.
        for key, value in headers.items():
            if isinstance(value, six.text_type):
                # We have to do this because request.headers is not
                # normal dictionary.  It has the (unintuitive) behavior
                # of aggregating repeated setattr calls for the same
                # key value.  For example:
                # headers['foo'] = 'a'; headers['foo'] = 'b'
                # list(headers) will print ['foo', 'foo'].
                del headers[key]
                headers[key] = value.encode('utf-8')

    def prepare_request(self, request):
        self._encode_headers(request.headers)
        return request.prepare()

    def _send_request(self, request_dict, operation_model, request_signer,
                      allow_redirects=True):
        attempts = 1
        request = self.create_request(request_dict, operation_model, request_signer)
        success_response, exception = self._get_response(
            request, operation_model, allow_redirects, attempts)
        while self._needs_retry(attempts, operation_model,
                                success_response, exception):
            attempts += 1
            # If there is a stream associated with the request, we need
            # to reset it before attempting to send the request again.
            # This will ensure that we resend the entire contents of the
            # body.
            request.reset_stream()
            # Create a new request when retried (including a new signature).
            request = self.create_request(
                request_dict,
                operation_model=operation_model,
                request_signer=request_signer)
            success_response, exception = self._get_response(
                request, operation_model, allow_redirects, attempts)
        if exception is not None:
            raise exception
        else:
            return success_response

    def _get_response(self, request, operation_model, allow_redirects, attempts):
        # This will return a tuple of (success_response, exception)
        # and success_response is itself a tuple of
        # (http_response, parsed_dict).
        # If an exception occurs then the success_response is None.
        # If no exception occurs then exception is None.
        try:
            http_response = self.http_session.send(
                request,
                verify=self.tls_verification,
                stream=False,
                proxies=self.proxies,
                timeout=self.timeout,
                allow_redirects=allow_redirects)
        except ConnectionError as e:
            # For a connection error, if it looks like it's a DNS
            # lookup issue, 99% of the time this is due to a misconfigured
            # region/endpoint so we'll raise a more specific error message
            # to help users.
            if self._looks_like_dns_error(e):
                endpoint_url = e.request.url
                better_exception = EndpointConnectionError(
                    endpoint_url=endpoint_url, error=e)
                return (None, better_exception)
            else:
                return (None, e)
        except Exception as e:
            return (None, e)

        if LOG.isEnabledFor(logging.DEBUG):
            skip_keys = [
                'content-security-policy',
                'x-content-type-options',
                'x-frame-options',
                'x-xss-protection'
            ]
            logged_headers = {k: v for k, v in dict(http_response.headers).items()
                              if k not in skip_keys}
            LOG.debug("Response headers: %s", pprint.pformat(logged_headers, indent=2))

        # This returns the http_response and the parsed_data.
        response_dict = convert_to_response_dict(http_response,
                                                 operation_model)
        parser = self._response_parser_factory.create_parser()
        return ((http_response, parser.parse(response_dict,
                                             operation_model.output_shape)),
                None)

    def _looks_like_dns_error(self, e):
        return 'gaierror' in str(e) and e.request is not None

    def _needs_retry(self,
                     attempts,
                     operation_model,
                     response=None,
                     caught_exception=None):
        if self._retry_handler is None:
            return False
        handler_response = \
            self._retry_handler(attempts=attempts,
                                response=response,
                                caught_exception=caught_exception)
        if handler_response is None:
            return False
        else:
            # Request needs to be retried, and we need to sleep
            # for the specified number of times.
            LOG.info("Response indicates retry is needed, sleeping for "
                     "%s seconds", handler_response)
            time.sleep(handler_response)
            return True



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from datetime import datetime
import decimal
import os
import sys

from cdpcli.model import ServiceModel
from cdpcli.model import ShapeResolver
from cdpcli.validate import ParamValidator
from tests import unittest
import yaml

MODEL_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'validate')


class BaseTestValidate(unittest.TestCase):

    def setUp(self):
        self.model = yaml.safe_load(open(os.path.join(MODEL_DIR, 'service.yaml')))
        self.service_model = ServiceModel(self.model, 'servicename')
        self.resolver = ShapeResolver(self.model['definitions'])

    def assert_has_validation_errors(self, shape, params, errors):
        errors_found = self.get_validation_error_message(shape, params)
        self.assertTrue(errors_found.has_errors())
        error_message = errors_found.generate_report()
        for error in errors:
            self.assertIn(error, error_message)

    def get_validation_error_message(self, shape, params):
        validator = ParamValidator()
        return validator.validate(params, shape)


class TestValidateRequiredParams(BaseTestValidate):

    def test_validate_required_params(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('director', 'Director'),
            {'name': 'foo'},
            ['Missing required parameter'])

    def test_validate_nested_required_param(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('director', 'Director'),
            {'name': 'foo',
             'crn': 'bar',
             'endpoint': {'address': 'baz'}},
            ['Missing required parameter'])

    def test_validate_unknown_param(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('director', 'Director'),
            {'name': 'foo',
             'crn': 'bar',
             'unknown': 'unknown'},
            ['Unknown parameter'])


class TestValidateTypes(BaseTestValidate):

    def test_validate_string(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'stringparam': 24,
             'intparam': 'notint',
             'boolparam': 'notbool',
             'numberparam': 'notnumber',
             'datetimeparam': 'notdatetime',
             'arrayparam': 'notarray',
             'objectparam': 'notobject',
             'mapparam': 'notmap',
             'blobparam': 123},
            ['Invalid type for parameter stringparam',
             'Invalid type for parameter intparam',
             'Invalid type for parameter boolparam',
             'Invalid type for parameter numberparam',
             'Invalid type for parameter datetimeparam',
             'Invalid type for parameter arrayparam',
             'Invalid type for parameter objectparam',
             'Invalid type for parameter mapparam',
             'Invalid type for parameter blobparam'])

    def test_datetime_type_accepts_datetime_object(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'datetimeparam': datetime.now()})
        self.assertEqual(errors.generate_report(), '')

    def test_datetime_accepts_string_timestamp(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'datetimeparam': '2014-01-01 12:00:00'})
        self.assertEqual(errors.generate_report(), '')

    def test_can_handle_none_datetimes(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'datetimeparam': None},
            ['Invalid type for parameter datetimeparam'])

    def test_valid_boolean(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'boolparam': True})
        self.assertEqual(errors.generate_report(), '')

    def test_valid_map(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'mapparam': {'key1': 'value1', 'key2': 'value2'}})
        self.assertEqual(errors.generate_report(), '')

    def test_invalid_key_map(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'mapparam': {'key1': 'value1', 20: 'value2'}},
            ["Invalid type for parameter mapparam (key: 20), value: 20"])

    def test_invalid_value_map(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'mapparam': {'key1': 'value1', 'key2': 2}},
            ["Invalid type for parameter mapparam.key2, value: 2"])

    def test_valid_blob_base64(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'blobparam': 'SGVsbG8gV29ybGQ='})
        self.assertEqual(errors.generate_report(), '')

    def test_valid_blob_empty_base64(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'blobparam': ''})
        self.assertEqual(errors.generate_report(), '')

    def test_invalid_value_blob_base64(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'blobparam': 'ABC'},
            ["Invalid base64 value for parameter blobparam, error: Incorrect padding"])

    if sys.version_info.major >= 3:  # bytes is introduced in python3
        def test_valid_blob_bytes(self):
            errors = self.get_validation_error_message(
                self.resolver.get_shape_by_name('typestest', 'TypesTest'),
                {'blobparam': b'BlobTest'})
            self.assertEqual(errors.generate_report(), '')

    if sys.version_info.major >= 3:  # bytes is introduced in python3
        def test_valid_blob_empty_bytes(self):
            errors = self.get_validation_error_message(
                self.resolver.get_shape_by_name('typestest', 'TypesTest'),
                {'blobparam': b''})
            self.assertEqual(errors.generate_report(), '')

    def test_valid_blob_bytearray(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'blobparam': bytearray(b'BlobTest')})
        self.assertEqual(errors.generate_report(), '')

    def test_valid_blob_empty_bytearray(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('typestest', 'TypesTest'),
            {'blobparam': bytearray(b'')})
        self.assertEqual(errors.generate_report(), '')


class TestValidateRanges(BaseTestValidate):

    def test_less_than_range(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'intparam': -10,
             'numberparam': -10.0},
            ['Invalid range for parameter intparam',
             'Invalid range for parameter numberparam'])

    def test_more_than_range(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'intparam': 100000,
             'numberparam': 100000.0},
            ['Invalid range for parameter intparam',
             'Invalid range for parameter numberparam'])

    def test_within_range(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'intparam': 50,
             'numberparam': 50.0})
        self.assertEqual(errors.generate_report(), '')

    def test_string_min_length_contraint(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'stringparam': ''},
            ['Invalid length for parameter stringparam'])

    def test_string_max_length_contraint(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'stringparam': 'thisistoolong'},
            ['Invalid length for parameter stringparam'])

    def test_list_min_length_constraint(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'arrayparam': []},
            ['Invalid length for parameter arrayparam'])

    def test_list_max_length_constraint(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'arrayparam': ['a', 'b', 'c', 'd', 'e', 'f']},
            ['Invalid length for parameter arrayparam'])

    def test_only_min_value_specified(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'minonlystringparam': ''},
            ['Invalid length for parameter minonlystringparam'])

    def test_only_max_value_specified(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'maxonlystringparam': 'thisistoolong'},
            ['Invalid length for parameter maxonlystringparam'])

    if sys.version_info.major >= 3:  # bytes is introduced in python3
        def test_blob_min_length_contraint_bytes(self):
            self.assert_has_validation_errors(
                self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
                {'blobparam': b''},
                ['Invalid length for parameter blobparam'])

    if sys.version_info.major >= 3:  # bytes is introduced in python3
        def test_blob_max_length_contraint_bytes(self):
            self.assert_has_validation_errors(
                self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
                {'blobparam': b'thisistoolong'},
                ['Invalid length for parameter blobparam'])

    def test_blob_min_length_contraint_bytearray(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'blobparam': bytearray(b'')},
            ['Invalid length for parameter blobparam'])

    def test_blob_max_length_contraint_bytearray(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'blobparam': bytearray(b'thisistoolong')},
            ['Invalid length for parameter blobparam'])

    def test_blob_min_length_contraint_base64(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'blobparam': ''},
            ['Invalid length for parameter blobparam'])

    def test_blob_max_length_contraint_base64(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('rangetest', 'RangeTest'),
            {'blobparam': 'dGhpc2lzdG9vbG9uZw=='},
            ['Invalid length for parameter blobparam'])


class TestValidationNumberType(BaseTestValidate):

    def test_range_number(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('numbertest', 'NumberTest'),
            {'numberparam': 1},
            ['Invalid range for parameter numberparam'])

    def test_decimal_allowed(self):
        errors = self.get_validation_error_message(
            self.resolver.get_shape_by_name('numbertest', 'NumberTest'),
            {'numberparam': decimal.Decimal('22.12345')})
        self.assertEqual(errors.generate_report(), '')

    def test_decimal_still_validates_range(self):
        self.assert_has_validation_errors(
            self.resolver.get_shape_by_name('numbertest', 'NumberTest'),
            {'numberparam': decimal.Decimal('1')},
            ['Invalid range for parameter numberparam'])



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli.exceptions import DfExtensionError
from cdpcli.extensions.df import DfExtension, upload_workload_asset
from mock import Mock
from tests import unittest


BASE_DIR = os.path.dirname(os.path.abspath(__file__))


class TestDfExtension(unittest.TestCase):

    def setUp(self):
        self.client = Mock()
        self.client.raise_error.side_effect = Exception('ClientError')
        self.client_creator = Mock(return_value=self.client)
        self.parsed_globals = Mock(output=None)

    def test_df_upload_flow(self):
        body_list = []

        def _make_request(*args, **kwargs):
            body_list.append(args[4].read())
            return Mock(status_code=200), {}

        self.client.make_request.side_effect = _make_request

        operation_model = Mock()
        operation_model.service_model = Mock(service_name='df')
        operation_model.name = 'importFlowDefinition'
        operation_model.http = {
            'method': 'post',
            'requestUri': 'https://test.com/path/importFlowDefinition'}
        parameters = {
            'name': 'flow_name',
            'description': 'flow_description\nline 2',
            'comments': 'flow_comments',
            'file': os.path.join(BASE_DIR, 'df.flow.json')}
        df_extension = DfExtension()
        df_extension.invoke(self.client_creator, operation_model,
                            parameters, None, self.parsed_globals)

        self.assertEqual(1, self.client.make_request.call_count)
        args, kwargs = self.client.make_request.call_args
        self.assertEqual('importFlowDefinition', args[0])
        self.assertEqual('post', args[1])
        self.assertEqual('https://test.com/path/importFlowDefinition', args[2])
        self.assertEqual(
            {'Content-Type': 'application/json',
             'Flow-Definition-Name': 'flow_name',
             'Flow-Definition-Description': 'flow_description%0Aline%202',
             'Flow-Definition-Comments': 'flow_comments'},
            args[3])
        body = body_list[0]
        self.assertEqual(
            (b'{"flowContents":{"identifier":"97047f2a-9d0d-3669-977e-9d022308feb9",'
             b'"name":"Simple","comments":""},"parameterContexts":{},"flowEncoding'
             b'Version":"1.0"}'),
            body)

    def test_df_upload_flow_no_optional_parameters(self):
        body_list = []

        def _make_request(*args, **kwargs):
            body_list.append(args[4].read())
            return Mock(status_code=200), {}

        self.client.make_request.side_effect = _make_request

        operation_model = Mock()
        operation_model.service_model = Mock(service_name='df')
        operation_model.name = 'importFlowDefinition'
        operation_model.http = {
            'method': 'post',
            'requestUri': 'https://test.com/path/importFlowDefinition'}
        parameters = {
            'name': 'flow_name',
            'file': os.path.join(BASE_DIR, 'df.flow.json')}
        df_extension = DfExtension()
        df_extension.invoke(self.client_creator, operation_model,
                            parameters, None, self.parsed_globals)

        self.assertEqual(1, self.client.make_request.call_count)
        args, kwargs = self.client.make_request.call_args
        self.assertEqual('importFlowDefinition', args[0])
        self.assertEqual('post', args[1])
        self.assertEqual('https://test.com/path/importFlowDefinition', args[2])
        self.assertEqual(
            {'Content-Type': 'application/json',
             'Flow-Definition-Name': 'flow_name'},
            args[3])
        body = body_list[0]
        self.assertEqual(
            (b'{"flowContents":{"identifier":"97047f2a-9d0d-3669-977e-9d022308feb9",'
             b'"name":"Simple","comments":""},"parameterContexts":{},"flowEncoding'
             b'Version":"1.0"}'),
            body)

    def test_df_upload_flow_version(self):
        body_list = []

        def _make_request(*args, **kwargs):
            body_list.append(args[4].read())
            return Mock(status_code=200), {}

        self.client.make_request.side_effect = _make_request

        operation_model = Mock()
        operation_model.service_model = Mock(service_name='df')
        operation_model.name = 'importFlowDefinitionVersion'
        operation_model.http = {
            'method': 'post',
            'requestUri': 'https://test.com/path/importFlowDefinitionVersion'}
        parameters = {
            'flowCrn': 'flow_crn',
            'comments': 'flow_comments\non 2 lines',
            'file': os.path.join(BASE_DIR, 'df.flow.json')}
        df_extension = DfExtension()
        df_extension.invoke(self.client_creator, operation_model,
                            parameters, None, self.parsed_globals)

        self.assertEqual(1, self.client.make_request.call_count)
        args, kwargs = self.client.make_request.call_args
        self.assertEqual('importFlowDefinitionVersion', args[0])
        self.assertEqual('post', args[1])
        self.assertEqual('https://test.com/path/importFlowDefinitionVersion', args[2])
        self.assertEqual(
            {'Content-Type': 'application/json',
             'Flow-Definition-Comments': 'flow_comments%0Aon%202%20lines'},
            args[3])
        body = body_list[0]
        self.assertEqual(
            (b'{"flowContents":{"identifier":"97047f2a-9d0d-3669-977e-9d022308feb9",'
             b'"name":"Simple","comments":""},"parameterContexts":{},"flowEncoding'
             b'Version":"1.0"}'),
            body)

    def test_df_upload_flow_version_no_optional_parameters(self):
        body_list = []

        def _make_request(*args, **kwargs):
            body_list.append(args[4].read())
            return Mock(status_code=200), {}

        self.client.make_request.side_effect = _make_request

        operation_model = Mock()
        operation_model.service_model = Mock(service_name='df')
        operation_model.name = 'importFlowDefinitionVersion'
        operation_model.http = {
            'method': 'post',
            'requestUri': 'https://test.com/path/importFlowDefinitionVersion'}
        parameters = {
            'flowCrn': 'flow_crn',
            'file': os.path.join(BASE_DIR, 'df.flow.json')}
        df_extension = DfExtension()
        df_extension.invoke(self.client_creator, operation_model,
                            parameters, None, self.parsed_globals)

        self.assertEqual(1, self.client.make_request.call_count)
        args, kwargs = self.client.make_request.call_args
        self.assertEqual('importFlowDefinitionVersion', args[0])
        self.assertEqual('post', args[1])
        self.assertEqual('https://test.com/path/importFlowDefinitionVersion', args[2])
        self.assertEqual(
            {'Content-Type': 'application/json'},
            args[3])
        body = body_list[0]
        self.assertEqual(
            (b'{"flowContents":{"identifier":"97047f2a-9d0d-3669-977e-9d022308feb9",'
             b'"name":"Simple","comments":""},"parameterContexts":{},"flowEncoding'
             b'Version":"1.0"}'),
            body)

    def test_df_workload_upload_asset(self):
        body_list = []

        def _make_request(*args, **kwargs):
            body_list.append(args[4].read())
            return Mock(status_code=200), {}

        self.client.make_request.side_effect = _make_request

        operation_model = Mock()
        operation_model.service_model = Mock(service_name='dfworkload')
        operation_model.name = 'uploadAsset'
        operation_model.http = {
            'method': 'post',
            'requestUri': '/dfx/api/rpc-v1/deployments/uploadAsset'}
        parameters = {
            'environmentCrn': 'env_crn',
            'parameterGroup': 'param_group',
            'parameterName': 'param_name',
            'deploymentRequestCrn': 'deployment_request_crn',
            'deploymentName': 'deployment_name',
            'assetUpdateRequestCrn': 'asset_update_request_crn',
            'filePath': os.path.join(BASE_DIR, 'df-workload.asset.bin')}
        df_extension = DfExtension()
        df_extension.invoke(self.client_creator, operation_model,
                            parameters, None, self.parsed_globals)

        self.assertEqual(1, self.client.make_request.call_count)
        args, kwargs = self.client.make_request.call_args
        self.assertEqual('uploadAsset', args[0])
        self.assertEqual('post', args[1])
        self.assertEqual('/dfx/api/rpc-v1/deployments/upload-asset-content', args[2])
        self.assertEqual(
            {'Content-Type': 'application/octet-stream',
             'Deployment-Request-Crn': 'deployment_request_crn',
             'Deployment-Name': 'deployment_name',
             'Asset-Update-Request-Crn': 'asset_update_request_crn',
             'Parameter-Group': 'param_group',
             'Parameter-Name': 'param_name',
             'File-Path': os.path.join(BASE_DIR, 'df-workload.asset.bin')},
            args[3])
        body = body_list[0]
        self.assertIsNotNone(body)
        self.assertEqual(42, len(body))

    def test_df_workload_upload_asset_no_optional_parameters(self):
        body_list = []

        def _make_request(*args, **kwargs):
            body_list.append(args[4].read())
            return Mock(status_code=200), {}

        self.client.make_request.side_effect = _make_request

        operation_model = Mock()
        operation_model.service_model = Mock(service_name='dfworkload')
        operation_model.name = 'uploadAsset'
        operation_model.http = {
            'method': 'post',
            'requestUri': '/dfx/api/rpc-v1/deployments/uploadAsset'}
        parameters = {
            'environmentCrn': 'env_crn',
            'parameterGroup': 'param_group',
            'parameterName': 'param_name',
            'filePath': os.path.join(BASE_DIR, 'df-workload.asset.bin')}
        df_extension = DfExtension()
        df_extension.invoke(self.client_creator, operation_model,
                            parameters, None, self.parsed_globals)

        self.assertEqual(1, self.client.make_request.call_count)
        args, kwargs = self.client.make_request.call_args
        self.assertEqual('uploadAsset', args[0])
        self.assertEqual('post', args[1])
        self.assertEqual('/dfx/api/rpc-v1/deployments/upload-asset-content', args[2])
        self.assertEqual(
            {'Content-Type': 'application/octet-stream',
             'Parameter-Group': 'param_group',
             'Parameter-Name': 'param_name',
             'File-Path': os.path.join(BASE_DIR, 'df-workload.asset.bin')},
            args[3])
        body = body_list[0]
        self.assertIsNotNone(body)
        self.assertEqual(42, len(body))

    def test_operation_not_supported(self):
        operation_model = Mock()
        operation_model.service_model = Mock(service_name='foo')
        operation_model.name = 'importFlowDefinition'
        df_extension = DfExtension()
        with self.assertRaisesRegex(DfExtensionError,
                                    'The DF extension failed: '
                                    'The operation is not supported. '
                                    'Service name: foo, '
                                    'operation name: importFlowDefinition'):
            df_extension.invoke(self.client_creator, operation_model,
                                None, None, None)
        self.assertEqual(0, self.client.make_api_call.call_count)
        self.assertEqual(0, self.client.make_request.call_count)

    def test_upload_file_not_found(self):
        parameters = {
            'environmentCrn': 'env_crn',
            'parameterGroup': 'param_group',
            'parameterName': 'param_name',
            'filePath': os.path.join(BASE_DIR, 'file-not-found')
        }

        with self.assertRaises(DfExtensionError):
            upload_workload_asset(self.client, parameters)
        self.assertEqual(0, self.client.make_api_call.call_count)
        self.assertEqual(0, self.client.make_request.call_count)

    def test_kpis(self):
        operation_model = Mock()
        operation_model.name = 'updateDeployment'
        operation_model.service_model = Mock(service_name='dfworkload')

        frequencyUnit = {
            'id': 'MINUTES'
        }
        frequencyUnitUpdated = {
            'id': 'MINUTES',
            'label': 'Minutes',
            'abbreviation': 'm'
        }
        kpi = {
            'metricId': 'cpuUtilization',
            'alert': {
                'thresholdMoreThan': {
                    'unitId': 'percentage',
                    'value': 75.5
                },
                'frequencyTolerance': {
                    'unit': frequencyUnit,
                    'value': 2.5
                }
            }
        }
        parameters = {
            'kpis': [
                kpi
            ]
        }

        parsed_args = Mock()
        parsed_globals = Mock()

        extension = DfExtension()

        result = extension.invoke(self.client_creator,
                                  operation_model,
                                  parameters,
                                  parsed_args,
                                  parsed_globals)
        self.assertTrue(result)

        self.assertEquals(
            kpi['alert']['frequencyTolerance']['unit'],
            frequencyUnitUpdated
        )

    def test_no_asset_references(self):
        operation_model = Mock()
        operation_model.name = 'updateDeployment'
        operation_model.service_model = Mock(service_name='dfworkload')
        parameters = {
            'parameterGroups': [
                {
                    'parameters': [
                        {
                            'name': 'parameterName',
                            'value': 'parameterValue'
                        }
                    ]
                }
            ]
        }
        parsed_args = Mock()
        parsed_globals = Mock()

        extension = DfExtension()

        result = extension.invoke(self.client_creator,
                                  operation_model,
                                  parameters,
                                  parsed_args,
                                  parsed_globals)
        self.assertTrue(result)

    def test_asset_references_uploaded(self):
        operation_model = Mock()
        operation_model.name = 'updateDeployment'
        operation_model.service_model = Mock(service_name='dfworkload')

        deployment_crn = 'DEPLOYMENT_CRN'
        asset_reference_name = 'df-workload.asset.bin'
        asset_reference_file_path = os.path.join(BASE_DIR, asset_reference_name)
        parameter_group_name = 'TestParameterGroup'
        parameter_name = 'TestParameter'

        parameters = {
            'deploymentCrn': deployment_crn,
            'parameterGroups': [
                {
                    'name': parameter_group_name,
                    'parameters': [
                        {
                            'name': parameter_name,
                            'assetReferences': [
                                {
                                    'path': asset_reference_file_path
                                }
                            ]
                        }
                    ]
                }
            ]
        }
        parsed_args = Mock()
        parsed_globals = Mock()

        extension = DfExtension()

        asset_update_request_crn = 'ASSET_UPDATE_REQUEST_CRN'
        update_response = {
            'assetUpdateRequestCrn': asset_update_request_crn
        }
        self.client.create_asset_update_request.return_value = update_response

        self.client.make_request.return_value = (Mock(status_code=200), {})

        result = extension.invoke(self.client_creator,
                                  operation_model,
                                  parameters,
                                  parsed_args,
                                  parsed_globals)
        self.assertTrue(result)

        found_request_crn = parameters.get('assetUpdateRequestCrn', None)
        self.assertEquals(asset_update_request_crn, found_request_crn)

        self.assertEquals(
            {
                'Content-Type': 'application/octet-stream',
                'Asset-Update-Request-Crn': asset_update_request_crn,
                'File-Path': asset_reference_file_path,
                'Parameter-Group': parameter_group_name,
                'Parameter-Name': parameter_name
            },
            self.client.make_request.call_args[0][3]
        )



# Copyright (c) 2021 Cloudera, Inc. All rights reserved.

import enum
import re


class DeploymentType(enum.Enum):
    PRIVATE_CLOUD = 'private'
    PUBLIC_CLOUD = 'public'


class ClassifyDeployment:

    def __init__(self, endpoint_url=None):
        self._endpoint_url = endpoint_url
        # As mentioned in configure.py, Public Cloud users may leave endpoint_url entry
        # blank, hence clarifying empty endpoint_url as Public Cloud.
        # Private Cloud users must enter a valid URL.
        if not self._endpoint_url:
            self._deployment = DeploymentType.PUBLIC_CLOUD
        else:
            self._deployment = self._deduce_deployment()

    def set_endpoint_url(self, endpoint_url):
        # Update instance's URL
        if not endpoint_url:
            self._deployment = DeploymentType.PUBLIC_CLOUD
        elif endpoint_url != self._endpoint_url:
            self._endpoint_url = endpoint_url
            self._deployment = self._deduce_deployment()

    def get_deployment_type(self):
        # Returns DeploymentType as an enum constant
        return self._deployment

    def _deduce_deployment(self):
        # Following console/cdp-js/lib/appConfigBase.ts
        # Following regular expression gives a clear identification of a public cloud
        # URL including different environments (int, dev, stage, private stack, etc.)
        #
        # for eg. Private Stack - https://console.cdp-priv.mow-dev.cloudera.com
        #         service URL - https://%sapi.thunderhead-int.cloudera.com
        if re.match('(.+(altus|cdp|dev|int|stage).cloudera.com)'
                    '|(http://localhost:\\d+.*)', self._endpoint_url):
            return DeploymentType.PUBLIC_CLOUD
        else:
            return DeploymentType.PRIVATE_CLOUD

    # Make individual assert methods for each environment type
    def is_private_cloud(self):
        return self._deployment == DeploymentType.PRIVATE_CLOUD

    def is_public_cloud(self):
        return self._deployment == DeploymentType.PUBLIC_CLOUD



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json

from cdpcli import xform_name
from cdpcli.argprocess import uri_param
from cdpcli.arguments import BooleanArgument
from cdpcli.arguments import CLIArgument
from cdpcli.arguments import ListArgument
import mock
from tests import BaseCLIDriverTest
from tests import temporary_file


# These tests use real service types so that we can
# verify the real shapes of services.
class BaseArgProcessTest(BaseCLIDriverTest):

    def get_param_model(self, dotted_name):
        service_name, operation_name, param_name = dotted_name.split('.')
        service_model = self.driver.get_service_model(service_name)
        operation = service_model.operation_model(operation_name)
        input_shape = operation.input_shape
        required_arguments = input_shape.required_members
        is_required = param_name in required_arguments
        member_shape = input_shape.members[param_name]
        type_name = member_shape.type_name
        cli_arg_name = xform_name(param_name, '-')
        if type_name == 'boolean':
            cls = BooleanArgument
        elif type_name == 'list':
            cls = ListArgument
        else:
            cls = CLIArgument
        return cls(cli_arg_name, member_shape, mock.Mock(), is_required)

    def create_argument(self, argument_name=None):
        if argument_name is None:
            argument_name = 'foo'
        argument = mock.Mock()
        argument.name = argument_name
        argument.cli_name = "--" + argument_name
        return argument


class TestURIParams(BaseArgProcessTest):
    def test_uri_param(self):
        p = self.get_param_model('iam.getUser.userId')
        with temporary_file('r+') as f:
            json_argument = json.dumps([{"Name": "user-id", "Values": ["u-1234"]}])
            f.write(json_argument)
            f.flush()
            result = uri_param(p, 'file://%s' % f.name)
        self.assertEqual(result, json_argument)

    def test_uri_param_no_paramfile_false(self):
        p = self.get_param_model('iam.getUser.userId')
        p.no_paramfile = False
        with temporary_file('r+') as f:
            json_argument = json.dumps([{"Name": "user-id", "Values": ["u-1234"]}])
            f.write(json_argument)
            f.flush()
            result = uri_param(p, 'file://%s' % f.name)
        self.assertEqual(result, json_argument)

    def test_uri_param_no_paramfile_true(self):
        p = self.get_param_model('iam.getUser.userId')
        p.no_paramfile = True
        with temporary_file('r+') as f:
            json_argument = json.dumps([{"Name": "user-id", "Values": ["u-1234"]}])
            f.write(json_argument)
            f.flush()
            result = uri_param(p, 'file://%s' % f.name)
        self.assertEqual(result, None)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2017 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from cdpcli.doc.restdoc import ReSTDocument
from cdpcli.docs import OperationDocumentGenerator
from cdpcli.model import DenormalizedStructureBuilder
from cdpcli.model import ShapeResolver
from cdpcli.model import StringShape
import mock
from tests import unittest


class TestRecursiveShapes(unittest.TestCase):
    def setUp(self):
        self.arg_table = {}
        self.help_command = mock.Mock()
        self.help_command.arg_table = self.arg_table
        self.operation_model = mock.Mock()
        self.operation_model.service_model.operation_names = []
        self.help_command.obj = self.operation_model
        self.operation_generator = OperationDocumentGenerator(self.help_command)

    def assert_rendered_docs_contain(self, expected):
        writes = [args[0][0] for args in
                  self.help_command.doc.write.call_args_list]
        writes = '\n'.join(writes)
        self.assertIn(expected, writes)


class TestTranslationMap(unittest.TestCase):
    def setUp(self):
        self.arg_table = {}
        self.help_command = mock.Mock()
        self.help_command.arg_table = self.arg_table
        self.operation_model = mock.Mock()
        self.operation_model.service_model.operation_names = []
        self.help_command.obj = self.operation_model
        self.operation_generator = OperationDocumentGenerator(self.help_command)

    def assert_rendered_docs_contain(self, expected):
        writes = [args[0][0] for args in
                  self.help_command.doc.write.call_args_list]
        writes = '\n'.join(writes)
        self.assertIn(expected, writes)

    def test_boolean_arg_groups(self):
        builder = DenormalizedStructureBuilder()
        input_model = builder.with_members({
            'Flag': {'type': 'boolean'}
        }).build_model()
        argument_model = input_model.members['Flag']
        argument_model.name = 'Flag'
        self.arg_table['flag'] = mock.Mock(
            cli_type_name='boolean', argument_model=argument_model)
        self.arg_table['no-flag'] = mock.Mock(
            cli_type_name='boolean', argument_model=argument_model)
        # The --no-flag should not be used in the translation.
        self.assertEqual(
            self.operation_generator.build_translation_map(),
            {'Flag': 'flag'})


class TestCLIDocumentEventHandler(unittest.TestCase):
    def setUp(self):
        self.session = mock.Mock()
        self.obj = None
        self.command_table = {}
        self.arg_table = {}
        self.name = 'my-command'

    def create_help_command(self):
        help_command = mock.Mock()
        help_command.doc = ReSTDocument()
        help_command.arg_table = {}
        operation_model = mock.Mock()
        operation_model.documentation = 'description'
        operation_model.service_model.operation_names = []
        help_command.obj = operation_model
        return help_command

    def test_documents_enum_values(self):
        shape_map = {
            'EnumArg': {
                'type': 'string',
                'enum': ['FOO', 'BAZ']
            }
        }
        shape = StringShape('EnumArg',
                            shape_map['EnumArg'],
                            ShapeResolver(shape_map))
        arg_table = {'arg-name': mock.Mock(argument_model=shape,
                                           _UNDOCUMENTED=False)}
        help_command = mock.Mock()
        help_command.doc = ReSTDocument()
        help_command.arg_table = arg_table
        operation_model = mock.Mock()
        operation_model.service_model.operation_names = []
        help_command.obj = operation_model
        operation_generator = OperationDocumentGenerator(help_command)
        operation_generator.doc_option('arg-name', help_command)
        rendered = help_command.doc.getvalue().decode('utf-8')
        self.assertIn('Possible values', rendered)
        self.assertIn('FOO', rendered)
        self.assertIn('BAZ', rendered)



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import re

from ._version import get_versions


# Retrieve the version number from versioneer's code.
__version__ = get_versions()['version']
del get_versions

VERSION = __version__

CDPCLI_ROOT = os.path.dirname(os.path.abspath(__file__))

# Read in the release file and update the advertised version with it if it's
# not the default / PUBLIC release.
_release_file_path = os.path.normpath("{0}/data/release.txt".format(CDPCLI_ROOT))
with open(_release_file_path) as releaseFile:
    RELEASE = releaseFile.readline()

if RELEASE != 'PUBLIC':
    VERSION += ' (%s)' % RELEASE

# Used to specify anonymous (unsigned) request signature
UNSIGNED = object()


SCALAR_TYPES = set(['string',
                    'float',
                    'integer',
                    'long',
                    'boolean',
                    'double',
                    'blob',
                    'datetime'])

LIST_TYPE = 'array'
OBJECT_TYPE = 'object'
MAP_TYPE = 'map'
REF_KEY = '$ref'
TYPE_KEY = 'type'
REF_NAME_PREFIX = '#/definitions/'
ADDITIONAL_PROPERTIES = 'additionalProperties'

COMPLEX_TYPES = set([OBJECT_TYPE,
                     LIST_TYPE,
                     MAP_TYPE])

DEFAULT_PROFILE_NAME = 'default'
CDP_ACCESS_KEY_ID_KEY_NAME = 'cdp_access_key_id'
CDP_PRIVATE_KEY_KEY_NAME = 'cdp_private_key'
CDP_ACCESS_TOKEN_KEY_NAME = 'cdp_access_token'
CDP_REGION_KEY_NAME = 'cdp_region'
CDP_ENDPOINT_URL_KEY_NAME = 'cdp_endpoint_url'
ENDPOINT_URL_KEY_NAME = 'endpoint_url'
FORM_FACTOR_KEY_NAME = 'form_factor'
# Python argparse has a bug in which '-' are not parsed correctly if they appear
# as values for other arguments, see: http://bugs.python.org/issue9334 for more
# details. This defines special encoding for dash that we will "decode" and
# replace with a dash. The reason we are using \\ is that there is a non zero
# chance that customers can discover this themselves.
ARGPARSE_DASH_ENCODING = '\\-'

# Prepopulate the cache with special cases that don't match our regular
# transformation.
_xform_cache = {
    ('s3GuardConfiguration', '-'): 's3-guard-configuration',
    ('s3GuardTableName', '-'): 's3-guard-table-name',
}
_first_cap_regex = re.compile('(.)([A-Z][a-z]+)')
_number_cap_regex = re.compile('([a-z])([0-9]+)')
_end_cap_regex = re.compile('([a-z0-9])([A-Z])')
# The regex below handles the special case where some acryonym
# name is pluralized, e.g GatewayARNs, ListWebACLs, SomeCNAMEs.
_special_case_transform = re.compile('[A-Z]{3,}s$')


def xform_name(name, sep='_', _xform_cache=_xform_cache):
    if sep in name:
        # If the sep is in the name, assume that it's already
        # transformed and return the string unchanged.
        return name
    key = (name, sep)
    if key not in _xform_cache:
        if _special_case_transform.search(name) is not None:
            is_special = _special_case_transform.search(name)
            matched = is_special.group()
            # Replace something like CRNs, ACLs with _crns, _acls.
            name = name[:-len(matched)] + sep + matched.lower()
        s1 = _first_cap_regex.sub(r'\1' + sep + r'\2', name)
        s2 = _number_cap_regex.sub(r'\1' + sep + r'\2', s1)
        transformed = _end_cap_regex.sub(r'\1' + sep + r'\2', s2).lower()
        _xform_cache[key] = transformed
    return _xform_cache[key]



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

NOT_SET = '<not set>'
PREDEFINED_SECTION_NAMES = ('preview',)
CREDENTIAL_FILE_COMMENT = """Note on private key format. We expect the private key to be in a modified PEM
format in which newlines are replaced with \\n."""


class ConfigValue(object):

    def __init__(self, value, source_type, source):
        self.value = value
        self.source_type = source_type
        self.source = source

    def mask_value(self):
        if self.value is NOT_SET:
            return
        self.value = mask_value(self.value)


def mask_value(current_value):
    if current_value is None:
        return 'None'
    else:
        return ('*' * 16) + current_value[-4:]



# Copyright 2012-2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Modifications made by Cloudera are:
#     Copyright (c) 2016 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import copy
import os
import sys

from cdpcli.thirdparty import six  # noqa
from dateutil.tz import tzlocal, tzwinlocal

if six.PY3:
    from base64 import encodebytes  # noqa
    from email.utils import formatdate  # noqa
    from http.client import HTTPResponse  # noqa
    import locale
    from six.moves import http_client
    from urllib.parse import urljoin  # noqa
    from urllib.parse import urlsplit  # noqa
    from urllib.parse import urlunsplit  # noqa

    raw_input = input

    class HTTPHeaders(http_client.HTTPMessage):
        pass

    def get_stdout_text_writer():
        return sys.stdout

    def ensure_unicode(s, encoding=None, errors=None):
        # NOOP in Python 3, because every string is already unicode
        return s

    def compat_open(filename, mode='r', encoding=None):
        """Back-port open() that accepts an encoding argument.

        In python3 this uses the built in open() and in python2 this
        uses the io.open() function.

        If the file is not being opened in binary mode, then we'll
        use locale.getpreferredencoding() to find the preferred
        encoding.

        """
        if 'b' not in mode:
            encoding = locale.getpreferredencoding()
        return open(filename, mode, encoding=encoding)

else:
    from base64 import encodestring as encodebytes  # noqa
    import codecs
    from email.message import Message
    from email.Utils import formatdate  # noqa
    from httplib import HTTPResponse  # noqa
    import io
    import locale
    from urlparse import urljoin  # noqa
    from urlparse import urlsplit  # noqa
    from urlparse import urlunsplit  # noqa

    raw_input = raw_input

    class HTTPHeaders(Message):

        # The __iter__ method is not available in python2.x, so we have
        # to port the py3 version.
        def __iter__(self):
            for field, value in self._headers:
                yield field

    def get_stdout_text_writer():
        # In python3, all the sys.stdout/sys.stderr streams are in text
        # mode.  This means they expect unicode, and will encode the
        # unicode automatically before actually writing to stdout/stderr.
        # In python2, that's not the case.  In order to provide a consistent
        # interface, we can create a wrapper around sys.stdout that will take
        # unicode, and automatically encode it to the preferred encoding.
        # That way consumers can just call get_stdout_text_writer() and write
        # unicode to the returned stream.  Note that get_stdout_text_writer
        # just returns sys.stdout in the PY3 section above because python3
        # handles this.
        return codecs.getwriter(locale.getpreferredencoding())(sys.stdout)

    def ensure_unicode(s, encoding='utf-8', errors='strict'):
        if s is None:
            return None
        if isinstance(s, six.text_type):
            return s
        return unicode(s, encoding, errors)  # noqa

    def compat_open(filename, mode='r', encoding=None):
        # See docstring for compat_open in the PY3 section above.
        if 'b' not in mode:
            encoding = locale.getpreferredencoding()
        return io.open(filename, mode, encoding=encoding)

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict  # noqa

if sys.version_info[:2] == (2, 6):
    import simplejson as json
else:
    import json  # noqa


@classmethod
def from_dict(cls, d):
    new_instance = cls()
    for key, value in d.items():
        new_instance[key] = value
    return new_instance


@classmethod
def from_pairs(cls, pairs):
    new_instance = cls()
    for key, value in pairs:
        new_instance[key] = value
    return new_instance


HTTPHeaders.from_dict = from_dict
HTTPHeaders.from_pairs = from_pairs


def copy_kwargs(kwargs):
    """
    There is a bug in Python versions < 2.6.5 that prevents you from passing
    unicode keyword args (#4978).  This function takes a dictionary of kwargs and
    returns a copy.  If you are using Python < 2.6.5, it also encodes the keys to
    avoid this bug. Oh, and version_info wasn't a namedtuple back then, either!
    """
    vi = sys.version_info
    if vi[0] == 2 and vi[1] <= 6 and vi[3] < 5:
        copy_kwargs = {}
        for key in kwargs:
            copy_kwargs[key.encode('utf-8')] = kwargs[key]
    else:
        copy_kwargs = copy.copy(kwargs)
    return copy_kwargs


def compat_input(prompt, interactive_long_input=False):
    """
    Cygwin's pty's are based on pipes. Therefore, when it interacts with a Win32
    program (such as Win32 python), what that program sees is a pipe instead of
    a console. This is important because python buffers pipes, and so on a
    pty-based terminal, text will not necessarily appear immediately. In most
    cases, this isn't a big deal. But when we're doing an interactive prompt,
    the result is that the prompts won't display until we fill the buffer. Since
    raw_input does not flush the prompt, we need to manually write and flush it.

    See https://github.com/mintty/mintty/issues/56 for more details.
    """
    is_windows = sys.platform.startswith('win')
    if interactive_long_input:
        # See THUN-222 for context on why this is necessary
        if is_windows is False:
            os.system('stty -icanon')
    sys.stdout.write(prompt)
    try:
        sys.stdout.flush()
        return raw_input()
    finally:
        if is_windows is False:
            os.system('stty sane')


def compat_tzlocal():
    # https://github.com/dateutil/dateutil/issues/197
    # on windows to avoid this error-
    # return time.localtime(timestamp + time.timezone).tm_isdst
    # ValueError: (22, 'Invalid argument') use
    # tzwinlocal in place of tzlocal
    is_windows = sys.platform.startswith('win')
    if is_windows:
        return tzwinlocal()
    else:
        return tzlocal()



# Copyright 2021 Cloudera, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

from cdpcli import VERSION
from cdpcli.doc.restdoc import ReSTDocument
from cdpcli.extensions.commands import BasicCommand
from cdpcli.help import HelpCommand
from cdpcli.help import NullRenderer


class RefdocCommand(BasicCommand):
    """
    The 'cdp refdoc' command handler.
    """

    NAME = 'refdoc'
    DESCRIPTION = ('Generate reference documentation. Files formatted as '
                   'reStructuredText are written to the output directory. They '
                   'may be used as input to the Sphinx documentation generator '
                   'to produce a readable documentation site in any of several '
                   'formats.')
    EXAMPLES = (
        'To generate reference documentation::\n'
        '\n'
        '    $ cdp refdoc --output-directory sphinx/source\n'
    )
    SUBCOMMANDS = []
    ARG_TABLE = [
        {
            'name': 'output-directory',
            'help_text': 'Directory where generated files are written.',
            'action': 'store',
            'required': True,
            'cli_type_name': 'string'
        }
    ]

    def __init__(self):
        super(RefdocCommand, self).__init__()

    def _write_ref_docs(self, command, output_dir, client_creator, parsed_args,
                        parsed_globals):
        """
        Write all of the documents for one command. If the command has
        subcommands, then documents for those are generated recursively.
        """
        self._write_ref_doc(command, output_dir, client_creator, parsed_args,
                            parsed_globals)
        if hasattr(command, '_get_command_table'):
            subcommand_table = command._get_command_table()
        elif hasattr(command, 'subcommand_table'):
            subcommand_table = command.subcommand_table
        else:
            subcommand_table = None
        if subcommand_table:
            for subcommand in subcommand_table.values():
                self._write_ref_docs(subcommand, output_dir, client_creator,
                                     parsed_args, parsed_globals)

    def _write_ref_doc(self, command, output_dir, client_creator, parsed_args,
                       parsed_globals):
        """
        Write the document for a single command.
        """

        # Get the help command for the command to be documented.
        if isinstance(command, HelpCommand):
            help_command = command
        else:
            help_command = command.create_help_command()

        # Derive the path of the command's documentation file, relative to the
        # output directory.
        # - Top-level commands, and any other commands that have subcommands,
        #   get index.rst files in directories of their own. Examples:
        #   - command foo => foo/index.rst (whether it has subcommands or not)
        #   - subcommand bar, which itself has subcommands, of command foo =>
        #     foo/bar/index.rst
        # - Leaf subcommands get files named after themselves inside the
        #   directory of their parent command. Example:
        #   - subcommand baz of command foo => foo/baz.rst
        # - Commands with no lineage (just cdp!) go to index.rst. If there are
        #   ever multiple lineage-less commands, this needs to be changed!
        if hasattr(command, 'lineage'):
            lineage = command.lineage
        else:
            lineage = []
        if len(lineage) > 0:
            ref_doc_partial_path = '/'.join(map(lambda c: c.name, lineage))
            if (hasattr(command, '_get_command_table') and  # has subcommands
                command._get_command_table()) or \
               (hasattr(command, 'subcommand_table') and  # has subcommands
                command.subcommand_table) or \
               len(lineage) == 1:  # is a top-level command (force this case)
                ref_doc_path = '{}/index.rst'.format(ref_doc_partial_path)
            else:
                ref_doc_path = '{}.rst'.format(ref_doc_partial_path)
        else:
            ref_doc_path = 'index.rst'  # the cdp command itself
        print('- {}'.format(ref_doc_path))

        # Run the help command to generate a ReST document suitable for HTML
        # conversion.
        help_command.doc = ReSTDocument(target='html')
        help_command.renderer = NullRenderer()
        help_command.include_man_fields = False
        help_command(client_creator, [], parsed_globals)

        # Create the directory for the document, if it doesn't already exist.
        ref_doc_dir = os.path.join(output_dir, os.path.dirname(ref_doc_path))
        if not os.path.exists(ref_doc_dir):
            os.makedirs(ref_doc_dir)

        # Write out the doc "value", which is the ReST string, to the file.
        content = help_command.doc.getvalue()
        with open(os.path.join(output_dir, ref_doc_path), 'w') as ref_doc:
            ref_doc.write(str(content, encoding='utf-8'))

    def _run_main(self, client_creator, parsed_args, parsed_globals):
        if not os.path.exists(parsed_args.output_directory):
            os.makedirs(parsed_args.output_directory)

        # Write documents for every known command.
        command_table = parsed_globals.command_table  # uses argparser hack
        for command in command_table.values():
            self._write_ref_docs(command, parsed_args.output_directory,
                                 client_creator, parsed_args, parsed_globals)

        # Write a file with the version string. This may be passed as a Sphinx
        # configuration value for use in generated documentation.
        with open(os.path.join(parsed_args.output_directory,
                               '_version'), 'w') as version_file:
            version_file.write(VERSION)
